DUMMY_DEPLOYMENT_JSON = {
    'id': '1fd7acec-817f-4ee4-b0cc-7046a06744e7:35be715e-e8e9-11ec-8e84-93331d4150d4',
    'organization': '35be715e-e8e9-11ec-8e84-93331d4150d4',
    'name': 'WOrds',
    'envId': '008c3d5b-49f4-44d8-ba40-c906854b3829:35be715e-e8e9-11ec-8e84-93331d4150d4',
    'environment': '008c3d5b-49f4-44d8-ba40-c906854b3829:35be715e-e8e9-11ec-8e84-93331d4150d4',
    'engineVersion': "5.9.0",
    'engineType': "DC",
    'engineConfiguration': {"engineType": "DC", "engineVersion": "5.9.0", "engineVersionId": "DC:5.10.0::RC3"},
    'type': 'SELF',
    'desiredInstances': 1,
    'scalaBinaryVersion': "2.12",
    'state': 'ENABLED',
    'status': 'OK',
    'stateDisplayLabel': 'ACTIVE',
    'statusDetail': '',
    'statusLastUpdated': 1702313113931,
    'creator': '2cfced39-e8e9-11ec-8e84-d7b86be914f5@35be715e-e8e9-11ec-8e84-93331d4150d4',
    'createTime': 1695922045569,
    'lastModifiedBy': '2cfced39-e8e9-11ec-8e84-d7b86be914f5@35be715e-e8e9-11ec-8e84-93331d4150d4',
    'lastModifiedOn': 1695922788328,
    'deploymentTags': None,
    'rawDeploymentTags': ['self-managed-tag'],
    'locked': False,
    'lockReference': None,
    'lockChangedAt': None,
    'lockChangedBy': None,
    'engineShutdownTimeout': 10,
    'engineShutdownTimeoutTimestamp': None,
    'showClonedConfigDiff': None,
    'installType': 'TARBALL',
}

DUMMY_DEPLOYMENT_JSON_2 = {
    'id': '1fd7acec-817f-4ee4-b0cc-7046a06744e7:35be715e-e8e9-11ec-8e84-93331d4150d4',
    'organization': '35be715e-e8e9-11ec-8e84-93331d4150d4',
    'name': "Don't Update",
    'envId': '008c3d5b-49f4-44d8-ba40-c906854b3829:35be715e-e8e9-11ec-8e84-93331d4150d4',
    'environment': '008c3d5b-49f4-44d8-ba40-c906854b3829:35be715e-e8e9-11ec-8e84-93331d4150d4',
    'engineVersion': "5.9.0",
    'engineType': "DC",
    'engineConfiguration': {
        "engineType": "DC",
        "engineVersion": "5.9.0",
        "engineVersionId": "DC:5.10.0::RC3",
        'jvmConfig': {"a": "a"},
        'stageLibs': None,
        'labels': None,
        'externalResourcesUri': None,
    },
    'type': 'SELF',
    'desiredInstances': 1,
    'scalaBinaryVersion': "2.12",
    'state': 'ENABLED',
    'status': 'OK',
    'stateDisplayLabel': 'ACTIVE',
    'statusDetail': '',
    'statusLastUpdated': 1702313113931,
    'creator': '2cfced39-e8e9-11ec-8e84-d7b86be914f5@35be715e-e8e9-11ec-8e84-93331d4150d4',
    'createTime': 1695922045569,
    'lastModifiedBy': '2cfced39-e8e9-11ec-8e84-d7b86be914f5@35be715e-e8e9-11ec-8e84-93331d4150d4',
    'lastModifiedOn': 1695922788328,
    'deploymentTags': None,
    'rawDeploymentTags': ['self-managed-tag'],
    'locked': False,
    'lockReference': None,
    'lockChangedAt': None,
    'lockChangedBy': None,
    'engineShutdownTimeout': 10,
    'engineShutdownTimeoutTimestamp': None,
    'showClonedConfigDiff': None,
    'installType': 'TARBALL',
}
