from streamsets.sdk.utils import get_random_string

GET_ENGINE_RESPONSE_COLLECTOR_JSON = {
    'id': get_random_string(),
    'organization': '1a1b7cf4-862e-11ed-94d5-41a005fd7967',
    'httpUrl': 'http://f8728e856446:18630',
    'version': '5.6.0',
    'labels': [],
    'reportedLabels': ['deployment'],
    'pipelinesCommitted': [],
    'lastReportedTime': 1697835654564,
    'startUpTime': 1696872736374,
    'offsetProtocolVersion': 2,
    'edge': False,
    'cpuLoad': 0.7079236552920763,
    'totalMemory': 16606822400.0,
    'usedMemory': 581667472.0,
    'memoryUsedPercentage': 0.03502581396908297,
    'pipelinesCount': 0,
    'responding': True,
    'idSeparator': '__',
    'executorType': 'COLLECTOR',
    'maxCpuLoad': 80.0,
    'maxMemoryUsed': 100.0,
    'maxPipelinesRunning': 1000000,
    'cspDeploymentId': '14dc3111-b89d-488a-b9c1-4c6d7a767c39:1a1b7cf4-862e-11ed-94d5-41a005fd7967',
    'deploymentId': None,
    'buildTime': 1687277400000,
    'overrideEngineMaxes': False,
    'lastShutdownTimestamp': None,
}

GET_ENGINE_RESPONSE_TRANSFORMER_JSON = {
    'id': get_random_string(),
    'organization': '1a1b7cf4-862e-11ed-94d5-41a005fd7967',
    'httpUrl': 'http://f8728e856446:18630',
    'version': '5.6.0',
    'labels': [],
    'reportedLabels': ['deployment'],
    'pipelinesCommitted': [],
    'lastReportedTime': 1697835654564,
    'startUpTime': 1696872736374,
    'offsetProtocolVersion': 2,
    'edge': False,
    'cpuLoad': 0.7079236552920763,
    'totalMemory': 16606822400.0,
    'usedMemory': 581667472.0,
    'memoryUsedPercentage': 0.03502581396908297,
    'pipelinesCount': 0,
    'responding': True,
    'idSeparator': '__',
    'executorType': 'TRANSFORMER',
    'maxCpuLoad': 80.0,
    'maxMemoryUsed': 100.0,
    'maxPipelinesRunning': 1000000,
    'cspDeploymentId': '14dc3111-b89d-488a-b9c1-4c6d7a767c39:1a1b7cf4-862e-11ed-94d5-41a005fd7967',
    'deploymentId': None,
    'buildTime': 1687277400000,
    'overrideEngineMaxes': False,
    'lastShutdownTimestamp': None,
}
