/* StreamSets Python SDK CSS file */

/* FONTS
------------------------------------------------------- */
@import url(https://fonts.googleapis.com/css?family=Lato:300,400,700,900);
@import url(https://fonts.googleapis.com/css?family=Montserrat:400,700);


/* GENERAL
------------------------------------------------------- */
body {
	font-family: Verdana, Geneva, sans-serif;
	font-size: 14px;
	line-height: 1.5em;
}

h1 {
	font-family: 'Lato', Helvetica, Arial, Lucida, sans-serif;
	font-size: 40px;
	line-height: 45px;
	margin-top: 45px ;
	margin-bottom: 0px;
	color: #003399 !important;
}

h2 {
	font-family: 'Lato', Helvetica, Arial, Lucida, sans-serif;
	font-size: 28px;
	margin-bottom: 10px;
	color: #003399 !important;
	border-bottom: 1px solid #e1e1e1;
}

h3 {
	font-family: 'Lato', Helvetica, Arial, Lucida, sans-serif;
	font-size: 26px;
	margin-bottom: 10px;
	color: #003399 !important;
	font-variant-caps: all-small-caps;
}

h4 {
	font-family: 'Lato', Helvetica, Arial, Lucida, sans-serif;
	font-size: 22px;
	margin-bottom: 10px;
	color: #003399 !important;
	font-variant-caps: all-small-caps;
}

h2::before {
    content: ">";
    color: #e0e0eb;
}

#site-navigation h1.site-logo {
    border-bottom: none;
}

#site-navigation {
    background-color: #eeeeee;
}

.bd-sidebar .nav {
    font-family: Verdana, Geneva, sans-serif;
    font-size: 14px;
    color: #555 !important;
}